import tempfile
import yaml

from assemblyline.common.identify import CUSTOM_URI_ID


def make_uri_file(directory: str, uri: str, params=None) -> str:
    with tempfile.NamedTemporaryFile(dir=directory, delete=False, mode="w") as out:
        out.write(CUSTOM_URI_ID)
        yaml.dump({"uri": uri}, out)
        if params:
            yaml.dump(params, out)
    return out.name


def normalize_uri_file(directory: str, filename: str) -> str:
    with open(filename, "r") as f:
        data = yaml.safe_load(f)
    uri = data.pop("uri")
    return make_uri_file(directory, uri, data)



from assemblyline import odm


@odm.model(index=True, store=True, description="Model for Empty Results")
class EmptyResult(odm.Model):
    # Empty results are gonna be an abstract construct
    #  Only a record of the key is saved for caching purposes
    expiry_ts = odm.Date(store=False, description="Expiry timestamp")



from assemblyline import odm
from assemblyline.common.attack_map import attack_map
from assemblyline.odm.models.ontology.results.network import REQUEST_METHODS, LOOKUP_TYPES

# Based on model in MaCo framework
CATEGORIES = ["adware", "apt", "backdoor", "banker", "bootkit", "bot", "browser_hijacker", "bruteforcer",
              "clickfraud", "cryptominer", "ddos", "downloader", "dropper", "exploitkit", "fakeav", "hacktool",
              "infostealer", "keylogger", "loader", "obfuscator", "pos", "proxy", "rat", "ransomware",
              "reverse_proxy", "rootkit", "scanner", "scareware", "spammer", "trojan", "virus", "wiper",
              "webshell", "worm"]
CONNECTION_USAGE = ["c2", "upload", "download", "propagate", "tunnel", "other", "ransom", "decoy"]


@odm.model(description="Encryption details")
class Encryption(odm.Model):
    algorithm = odm.Optional(odm.Text(), description="Algorithm")
    public_key = odm.Optional(odm.Text(), description="Public Key")
    key = odm.Optional(odm.Text(), description="Key")
    provider = odm.Optional(odm.Text(), description="Provider")
    mode = odm.Optional(odm.Text(), description="Mode")
    iv = odm.Optional(odm.Text(), description="Initialization Vector")
    seed = odm.Optional(odm.Text(), description="Seed")
    nonce = odm.Optional(odm.Text(), description="Nonce value")
    constants = odm.Optional(odm.List(odm.Text()), description="Constants")
    usage = odm.Optional(odm.Enum(values=["config", "communication", "binary", "ransom", "other"]),
                         description="Purpose of encryptions")


@odm.model(description="Binary data extracted by decoder")
class Binary(odm.Model):
    datatype = odm.Optional(odm.Enum(values=["payload", "config", "other"]))
    data = odm.Text()
    other = odm.Optional(odm.Mapping(odm.Any()), description="Other information")
    encryption = odm.Optional(odm.List(odm.Compound(Encryption)))


@odm.model(description="Usage of FTP connection")
class FTP(odm.Model):
    username = odm.Optional(odm.Text(), description="Username")
    password = odm.Optional(odm.Text(), description="Password")
    hostname = odm.Optional(odm.Text(), description="FTP Host")
    port = odm.Optional(odm.Integer(), description="FTP Port")
    path = odm.Optional(odm.Text(), description="FTP Path")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of FTP connection")


@odm.model(description="Usage of SMTP connection")
class SMTP(odm.Model):
    username = odm.Optional(odm.Text(), description="Username")
    password = odm.Optional(odm.Text(), description="Password")
    hostname = odm.Optional(odm.Text(), description="SMTP Host")
    port = odm.Optional(odm.Integer(), description="SMTP Port")
    mail_to = odm.Optional(odm.List(odm.Text()), description="Sent to")
    mail_from = odm.Optional(odm.Text(), description="Sent from")
    subject = odm.Optional(odm.Text(), description="Subject")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of SMTP connection")


@odm.model(description="Usage of HTTP connection")
class HTTP(odm.Model):
    uri = odm.Optional(odm.URI(), description="URI")
    protocol = odm.Optional(odm.Enum(values=["http", "https"]), description="Protocol")
    username = odm.Optional(odm.Text(), description="Username")
    password = odm.Optional(odm.Text(), description="Password")
    hostname = odm.Optional(odm.Text(), description="HTTP server")
    port = odm.Optional(odm.Integer(), description="HTTP Port")
    path = odm.Optional(odm.URIPath(), description="URI Path")
    query = odm.Optional(odm.Text(), description="Query parameters")
    fragment = odm.Optional(odm.Text(), description="Fragment")
    user_agent = odm.Optional(odm.Text(), description="User Agent")
    method = odm.Optional(odm.Enum(values=REQUEST_METHODS), description="Method")
    headers = odm.Optional(odm.Mapping(odm.Text()), description="HTTP Headers")
    max_size = odm.Optional(odm.Integer(), description="Maximum size")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of HTTP connection")


@odm.model(description="Usage of SSH connection")
class SSH(odm.Model):
    username = odm.Optional(odm.Text(), description="Username")
    password = odm.Optional(odm.Text(), description="Password")
    public_key = odm.Optional(odm.Text(), description="SSH Public Key")
    hostname = odm.Optional(odm.Text(), description="SSH Host")
    port = odm.Optional(odm.Integer(), description="SSH Port")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of SSH connection")


@odm.model(description="Usage of Proxy connection")
class Proxy(odm.Model):
    username = odm.Optional(odm.Text(), description="Username")
    password = odm.Optional(odm.Text(), description="Password")
    hostname = odm.Optional(odm.Text(), description="Proxy Host")
    port = odm.Optional(odm.Integer(), description="Proxy Port")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of proxy connection")
    protocol = odm.Optional(odm.Text(), description="Protocol used")


@odm.model(description="Usage of DNS connection")
class DNS(odm.Model):
    ip = odm.Optional(odm.IP(), description="IP of DNS server")
    port = odm.Optional(odm.Integer(), description="Port of DNS server")
    hostname = odm.Optional(odm.Text(), description="Hostname used in query")
    record_type = odm.Optional(odm.Enum(values=LOOKUP_TYPES), description="Type of DNS record")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of DNS connection")


@odm.model(description="Usage of General TCP/UDP connection")
class GeneralConnection(odm.Model):
    client_ip = odm.Optional(odm.IP(), description="Client IP")
    client_port = odm.Optional(odm.Integer(), description="Client Port")
    server_ip = odm.Optional(odm.IP(), description="Server IP")
    server_domain = odm.Optional(odm.Domain(), description="Server Domain")
    server_port = odm.Optional(odm.Integer(), description="Server Port")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of TCP/UDP connection")


@odm.model(description="Operating System services affected")
class Service(odm.Model):
    dll = odm.Optional(odm.Text(), description="DLL associated to service")
    name = odm.Optional(odm.Text(), description="Name of service")
    display_name = odm.Optional(odm.Text(), description="Display Name of service")
    description = odm.Optional(odm.Text(), description="Service Description")


@odm.model(description="Cryptocoin usage (ransomware/miner)")
class Cryptocurrency(odm.Model):
    coin = odm.Optional(odm.Text(), description="Name of coin used")
    address = odm.Optional(odm.Text(), description="Wallet address")
    random_amount = odm.Optional(odm.Integer(), description="Ransom amount")
    usage = odm.Optional(odm.Enum(values=['ransomware', 'miner', 'other']), description="Use of cryptocurrency")


@odm.model(description="File Paths")
class Path(odm.Model):
    path = odm.Optional(odm.Text(), description="Path")
    usage = odm.Optional(
        odm.Enum(values=['c2', 'config', 'install', 'plugins', 'logs', 'storage', 'other']),
        description="Use of path")


@odm.model(description="Inter-Process Communications")
class IPC(odm.Model):
    file = odm.Optional(odm.List(odm.Text()),
                        description="A record stored on disk, or a record synthesized on demand by a file server, "
                        "which can be accessed by multiple processes.")
    socket = odm.Optional(odm.List(odm.Text()),
                          description="Data sent over a network interface, either to a different process on the same "
                          "computer or to another computer on the network. Stream oriented (TCP; data written through a"
                          " socket requires formatting to preserve message boundaries) or more rarely message-oriented "
                          "(UDP, SCTP).")
    unix_domain_socket = odm.Optional(odm.List(odm.Text()),
                                      description="Similar to an internet socket, but all communication occurs within "
                                      "the kernel. Domain sockets use the file system as their address space. "
                                      "Processes reference a domain socket as an inode, and multiple processes can "
                                      "communicate with one socket.")
    memory_mapped_file = odm.Optional(odm.List(odm.Text()),
                                      description="A file mapped to RAM and can be modified by changing memory"
                                      "addresses directly instead of outputting to a stream. This shares the same "
                                      "benefits as a standard file.")
    message_queue = odm.Optional(odm.List(odm.Text()),
                                 description="A data stream similar to a socket, but which usually preserves message "
                                 "boundaries. Typically implemented by the operating system, they allow multiple "
                                 "processes to read and write to the message queue without being directly connected to "
                                 "each other.")
    anonymous_pipe = odm.Optional(odm.List(odm.Text()),
                                  description="A unidirectional data channel using standard input and output. Data "
                                  "written to the write-end of the pipe is buffered by the operating system until it "
                                  "is read from the read-end of the pipe. Two-way communication between processes can "
                                  'be achieved by using two pipes in opposite "directions".')
    named_pipe = odm.Optional(odm.List(odm.Text()),
                              description="A pipe that is treated like a file. Instead of using standard input and "
                              "output as with an anonymous pipe, processes write to and read from a named pipe, as if "
                              "it were a regular file.")
    process_names = odm.Optional(odm.List(odm.Text()),
                                 description="The process names involved in the IPC communication")
    shared_memory = odm.Optional(odm.Text(),
                                 description="Multiple processes are given access to the same block of memory, which "
                                 "creates a shared buffer for the processes to communicate with each other.")
    usage = odm.Optional(odm.Enum(values=CONNECTION_USAGE), description="Purpose of connection")


@odm.model(description="Registry")
class Registry(odm.Model):
    key = odm.Text(description="Registry key")
    value = odm.Optional(odm.Text(), description="Registry")
    usage = odm.Optional(
        odm.Enum(values=['persistence', 'store_data', 'store_payload', 'read', 'other']),
        description="Use of registry key")


@odm.model(description="Extracted Malware Configuration")
class MalwareConfig(odm.Model):
    config_extractor = odm.Keyword(description="Name of extractor")

    family = odm.List(odm.Text(), description="What family is this associated to?")
    version = odm.Optional(odm.Text(), description="Version of the malware")
    category = odm.Optional(odm.List(odm.Enum(values=CATEGORIES)), description="Category of malware")

    attack = odm.Optional(odm.List(odm.Enum(values=list(attack_map.keys()))), description="ATT&CK ID associated")
    capability_enabled = odm.Optional(odm.List(odm.Text()), description="Enabled Capabilities")
    capability_disabled = odm.Optional(odm.List(odm.Text()), description="Disabled Capabilities")
    campaign_id = odm.Optional(odm.List(odm.Text()), description="Campaign ID")
    identifier = odm.Optional(odm.List(odm.Text()), description="Identifier")
    decoded_strings = odm.Optional(odm.List(odm.Text()), description="Decoded Strings")
    password = odm.Optional(odm.List(odm.Text()), description="Passwords")
    mutex = odm.Optional(odm.List(odm.Text()), description="Mutex")
    pipe = odm.Optional(odm.List(odm.Text()), description="Pipe")
    ipc = odm.Optional(odm.List(odm.Compound(IPC)), description="IPC (similar to 'pipe' field but more detailed)")
    sleep_delay = odm.Optional(odm.Integer(), description="Sleep Delay")
    sleep_delay_jitter = odm.Optional(odm.Integer(), description="Sleep Delay Jitter")
    inject_exe = odm.Optional(odm.List(odm.Text()), description="Injected EXE")

    binaries = odm.Optional(odm.List(odm.Compound(Binary)), description="Binaries")
    ftp = odm.Optional(odm.List(odm.Compound(FTP)), description="FTPs")
    smtp = odm.Optional(odm.List(odm.Compound(SMTP)), description="SMTPs")
    http = odm.Optional(odm.List(odm.Compound(HTTP)), description="HTTPs")
    ssh = odm.Optional(odm.List(odm.Compound(SSH)), description="SSHs")
    proxy = odm.Optional(odm.List(odm.Compound(Proxy)), description="Proxies")
    dns = odm.Optional(odm.List(odm.Compound(DNS)), description="DNS")
    tcp = odm.Optional(odm.List(odm.Compound(GeneralConnection)), description="TCPs")
    udp = odm.Optional(odm.List(odm.Compound(GeneralConnection)), description="UDPs")
    encryption = odm.Optional(odm.List(odm.Compound(Encryption)), description="Encryptions")
    service = odm.Optional(odm.List(odm.Compound(Service)), description="Services")
    cryptocurrency = odm.Optional(odm.List(odm.Compound(Cryptocurrency)), description="Cryptocurrencies")
    paths = odm.Optional(odm.List(odm.Compound(Path)), description="Paths")
    registry = odm.Optional(odm.List(odm.Compound(Registry)), description="Registry")

    other = odm.Optional(odm.Mapping(odm.Any()), description="Other information")



from assemblyline import odm

REACTIONS_TYPES = {"thumbs_up", "thumbs_down", "love", "smile", "surprised", "party"}


@odm.model(index=True, store=True, description="""Detailed breakdown model of a file's Uniform Resource Identifier (URI).

URIInfo dissects a file's URI into its fundamental components, providing granular data for advanced search and identification. This includes the scheme, network location, path, and other elements such as query parameters and fragments. By parsing these components, Assemblyline allows for a more nuanced analysis of the source and context of a file, which is essential for forensic investigations and threat intelligence gathering.

Each of these descriptions aims to provide a clearer understanding of the purpose and utility of the respective models within Assemblyline, highlighting their roles in the broader context of malware analysis and cyber security operations.
""")
class URIInfo(odm.Model):
    uri: str = odm.Keyword(description="The complete Uniform Resource Identifier (URI) of the file.")

    # https://www.rfc-editor.org/rfc/rfc1808.html#section-2.1
    scheme: str = odm.Keyword(description="The scheme component of the URI (e.g., \"http\", \"ftp\").")
    netloc: str = odm.Keyword(description="The network location part of the URI, including the domain name and port.")
    path: str = odm.Optional(odm.Keyword(),description="The path component of the URI, specifying the resource within the host.")
    params: str = odm.Optional(odm.Keyword(), description="The parameters component of the URI, often used for session management.")
    query: str = odm.Optional(odm.Keyword(), description="The query string of the URI, containing data for server-side processing.")
    fragment: str = odm.Optional(odm.Keyword(), description="The fragment identifier of the URI, used to navigate to a specific part of the resource.")

    # Ease-of-use elements
    username: str = odm.Optional(odm.Keyword(), description="The username specified in the URI, if any.")
    password: str = odm.Optional(odm.Keyword(), description="The password specified in the URI, if any.")
    hostname: str = odm.Keyword(description="The hostname extracted from the netloc, representing the domain of the URI.")
    port: int = odm.Optional(odm.Integer(), description="The port number extracted from the netloc, representing the communication endpoint.")


@odm.model(index=True, store=True, description="""Tracking model for the occurrence and frequency of a file within the system.

The Seen model is designed to record and quantify the instances in which a file is encountered by Assemblyline. It keeps a count of the file's occurrences and logs the timestamps of the first and most recent sightings. This temporal information is crucial for understanding the prevalence and distribution of a file over time, aiding in threat trend analysis and situational awareness.
""")
class Seen(odm.Model):
    count = odm.Integer(default=1, description="The total number of times the file has been observed by the system.")
    first = odm.Date(default="NOW", description="The timestamp of the file's first sighting.")
    last = odm.Date(default="NOW", description="The timestamp of the file's most recent sighting.")


@odm.model(index=True, store=True, description="""Structured categorization model for labels applied to a file.

LabelCategories provide a systematic approach to classifying the characteristics and threat indicators of a file. This model organizes labels into distinct categories such as informational tags, technical techniques, and attribution links. By categorizing labels, analysts can efficiently navigate and assess the nature and potential threats associated with a file, streamlining the malware analysis process.
""")
class LabelCategories(odm.Model):
    info = odm.List(
        odm.Keyword(),
        description="Informational labels providing additional context about the file.", default=[])
    technique = odm.List(
        odm.Keyword(),
        description="An array of labels identifying the specific tactics, techniques, and procedures (TTPs) as defined by the MITRE ATT&CK® framework that are exhibited by the malware within the file. This field also includes labels for any detection signatures that triggered during analysis, providing insight into the malware's behavior and potential impact. Analysts can use these labels to correlate files with known adversary behavior and to enhance threat hunting and incident response activities.",
        default=[])
    attribution = odm.List(
        odm.Keyword(),
        description="Labels that relate to the attribution of the file, such as the associated threat actor or campaign.",
        default=[])


@odm.model(index=True, store=False, description="""Model that encapsulates user interactions with a comment.

The Reaction model captures the responses of users to comments made on a file. Reactions are simple expressions of agreement, disagreement, or sentiment, represented by a set of predefined icons. These reactions facilitate a quick, non-verbal form of feedback from users, enhancing collaborative analysis and engagement within the Assemblyline platform.
""")
class Reaction(odm.Model):
    icon = odm.Enum(values=REACTIONS_TYPES, description="Icon name representing the type of reaction given to a comment.")
    uname = odm.Keyword(description="The username of the individual who reacted to the comment.")


@odm.model(index=True, store=False, description="""Model that represents user annotations attached to a file.

A Comment is a user-generated note or observation that can be added to a file within Assemblyline. This feature enables analysts to record insights, share findings, and collaborate on the analysis of a file. Each comment is timestamped and associated with the username of the individual who authored it, creating an audit trail of analytical discourse.
""")
class Comment(odm.Model):
    cid = odm.UUID(description="Unique identifier for the comment.")
    uname = odm.Keyword(description="The username of the individual who authored the comment.")
    date = odm.Date(store=True, default="NOW", description="The date and time when the comment was posted.")
    text = odm.Text(description="The content of the comment as written by the user.")
    reactions = odm.List(odm.Compound(Reaction), default=[], description="An array of user reactions to the comment, such as likes or dislikes.")


@odm.model(index=True, store=True, description="""This section presents the detailed schema of the File object model within the Assemblyline application. Each entry in the schema represents a field that constitutes a File document within the File index. The information provided for each field includes the data type, a concise description, its requirement status, and the default value if any.

Understanding this schema is crucial for constructing effective and precise Lucene search queries. By leveraging the fields outlined in the table below, you can craft queries to retrieve specific information about files analyzed by Assemblyline. These fields are integral for in-depth data analysis, enabling you to filter and locate files based on various attributes such as type, hash values, classification, and many others.

Utilize this schema as a reference to enhance your search capabilities within the Assemblyline system, allowing for more targeted and refined data retrieval that aligns with your cybersecurity analysis needs.
""")
class File(odm.Model):
    archive_ts = odm.Optional(odm.Date(ai=False), description="Timestamp indicating when the file was archived.")
    ascii = odm.Keyword(index=False, store=False,
                        description="Provides a dotted ASCII representation of the first 64 bytes of the file.", ai=False)
    classification = odm.Classification(description="Security classification assigned to the file based on its contents and context.")
    comments = odm.List(odm.Compound(Comment), default=[], description="An array of user-generated comments pertaining to the file. See Comment model for more information.")
    entropy = odm.Float(description="A numerical value representing the file's entropy, which is defined as the level of randomness in the file's content, typically used to detect compression or encryption. High entropy may indicate obfuscation techniques such as encryption, commonly employed by malware to evade detection. This metric is not exclusive to malicious files, as legitimate files can also exhibit high entropy.")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Timestamp indicating when the file is scheduled to expire from the system.", ai=False)
    is_section_image = odm.Boolean(
        default=False, description="Indicates if the file is an image safe for web browser display, often part of analysis results.", ai=False)
    is_supplementary = odm.Boolean(default=False, description="Indicates if the file was created by an AssemblyLine service as supplementary data.", ai=False)
    hex = odm.Keyword(index=False, store=False, description="Hexadecimal representation of the first 64 bytes of the file.", ai=False)
    labels = odm.List(odm.Keyword(copyto="__text__"), description="Array of descriptive labels applied to the file for categorization and analysis.", default=[])
    label_categories = odm.Compound(LabelCategories, default={}, description="Structured categories for the labels applied to the file.", ai=False)
    md5 = odm.MD5(copyto="__text__", description="The MD5 hash of the file, used for identifying duplicates and verifying integrity.", ai=False)
    magic = odm.Keyword(store=False, description="Detailed file format information derived from an analysis using the libmagic library, including text descriptions of the file's content type and encoding.")
    mime = odm.Optional(odm.Keyword(store=False), description="The Multipurpose Internet Mail Extensions (MIME) type of the file as determined by libmagic, which identifies file types by checking their headers according to a predefined list of file types.")
    seen = odm.Compound(Seen, default={}, description="Records the frequency and timestamps of when the file was encountered.", ai=False)
    sha1 = odm.SHA1(copyto="__text__", description="The SHA1 hash of the file, providing a more secure alternative to MD5 for integrity checks.", ai=False)
    sha256 = odm.SHA256(copyto="__text__", description="The SHA256 hash of the file, offering a high level of security for integrity verification.")
    size = odm.Integer(description="Size of the file in bytes.")
    ssdeep = odm.SSDeepHash(store=False, description="The fuzzy hash of the file using SSDEEP, which is useful for identifying similar files.", ai=False)
    type = odm.Keyword(copyto="__text__", description="The file type as determined by the AssemblyLine file type identification service.")
    tlsh = odm.Optional(odm.Keyword(copyto="__text__"), description="A locality-sensitive hash (TLSH) of the file's content, useful for similarity comparisons.", ai=False)
    from_archive = odm.Boolean(index=False, store=False, default=False,
                               description="Indicates whether the file was retrieved from Assemblyline's archive during processing.", ai=False)
    uri_info = odm.Optional(odm.Compound(URIInfo), description="Detailed components of the file's URI for advanced search functionality.")



from __future__ import annotations
import hashlib
import json
import threading
import time
import typing

from collections import OrderedDict
from typing import Generic, TypeVar, Hashable, Tuple, Optional

import baseconv

from assemblyline.common.uid import get_random_id


if typing.TYPE_CHECKING:
    from assemblyline.odm.messages.task import Task


T = TypeVar('T')


class TimeExpiredCache(Generic[T]):
    """
    TimeExpiredCache is a thread safe caching object that will store any amount of items for
    a period of X seconds at maximum.

    A thread inside the cache is fired every "expiry_rate" seconds and will remove all items that
    meet their timeouts.

    If you add the same item twice, the second time you add the item it will be ignored or can
    raise an exception if specified. This will not freshen the timeout for the specified item.
    """

    def __init__(self, timeout: float, expiry_rate: float = 5, raise_on_error: bool = False):
        self.lock = threading.Lock()
        self.timeout = timeout
        self.expiry_rate = expiry_rate
        self.raise_on_error = raise_on_error
        self.cache: dict[Hashable, T] = {}
        self.timeout_list: list[Tuple[float, Hashable]] = []
        timeout_thread = threading.Thread(target=self._process_timeouts, name="_process_timeouts", daemon=True)
        timeout_thread.start()

    def __len__(self):
        with self.lock:
            return len(self.cache)

    def __str__(self):
        with self.lock:
            return 'TimeExpiredCache(%s): %s' % (self.timeout, str(self.cache.keys()))

    def _process_timeouts(self):
        while True:
            time.sleep(self.expiry_rate)
            current_time = time.time()
            index = 0

            with self.lock:
                for t, k in self.timeout_list:
                    if t >= current_time:
                        break

                    index += 1

                    self.cache.pop(k, None)

                self.timeout_list = self.timeout_list[index:]

    def add(self, key: Hashable, data: T):
        with self.lock:
            if key in self.cache:
                if self.raise_on_error:
                    raise KeyError("%s already in cache" % key)
                else:
                    return

            self.cache[key] = data
            self.timeout_list.append((time.time() + self.timeout, key))

    @typing.overload
    def get(self, key: Hashable) -> Optional[T]:
        ...

    @typing.overload
    def get(self, key: Hashable, default: T) -> T:
        ...

    def get(self, key: Hashable, default: Optional[T] = None) -> Optional[T]:
        with self.lock:
            return self.cache.get(key, default)

    def keys(self):
        with self.lock:
            return self.cache.keys()


class SizeExpiredCache(Generic[T]):
    """
    SizeExpiredCache is a thread safe caching object that will store only X number of item for
    caching at maximum.

    If more items are added, the oldest item is removed.

    If you add the same item twice, the second time you add the item it will be ignored or can
    raise an exception if specified. This will not freshen the item position in the cache.
    """

    def __init__(self, max_item_count: int, raise_on_error: bool = False):
        self.lock = threading.Lock()
        self.max_item_count = max_item_count
        self.cache: OrderedDict[Hashable, T] = OrderedDict()
        self.raise_on_error = raise_on_error

    def __len__(self):
        with self.lock:
            return len(self.cache)

    def __str__(self):
        with self.lock:
            return 'SizeExpiredCache(%s/%s): %s' % (len(self.cache), self.max_item_count, str(self.cache.keys()))

    def add(self, key: Hashable, data: T):
        with self.lock:
            if key in self.cache:
                if self.raise_on_error:
                    raise KeyError("%s already in cache" % key)
                else:
                    return

            self.cache[key] = data
            if len(self.cache) > self.max_item_count:
                self.cache.popitem(False)

    @typing.overload
    def get(self, key: Hashable) -> Optional[T]:
        ...

    @typing.overload
    def get(self, key: Hashable, default: T) -> T:
        ...

    def get(self, key: Hashable, default: Optional[T] = None) -> Optional[T]:
        with self.lock:
            return self.cache.get(key, default)

    def keys(self):
        with self.lock:
            return self.cache.keys()


def generate_conf_key(service_tool_version: Optional[str] = None, task: Optional[Task] = None, partial_result: bool = False) -> str:
    ignore_salt = None
    service_config = None
    submission_params_str = None

    if task is not None:
        service_config = json.dumps(sorted(task.service_config.items()))
        submission_params = {
            "deep_scan": task.deep_scan,
            "max_files": task.max_files,
            "min_classification": task.min_classification.value,
            "ignore_filtering": task.ignore_filtering,
        }
        submission_params_str = json.dumps(sorted(submission_params.items()))

        if task.ignore_cache or partial_result:
            ignore_salt = get_random_id()

    if service_tool_version is None and \
            service_config is None and \
            submission_params_str is None and \
            ignore_salt is None:
        return "0"

    total_str = f"{service_tool_version}_{service_config}_{submission_params_str}_{ignore_salt}".encode('utf-8')
    partial_md5 = hashlib.md5((str(total_str).encode('utf-8'))).hexdigest()[:16]
    return baseconv.base62.encode(int(partial_md5, 16))



from __future__ import annotations

from assemblyline import odm
from assemblyline.odm.models.service import FETCH_METHODS, SIGNATURE_DELIMITERS

REF_DEPENDENCY_CONFIG = "Refer to:<br>[Service - DependencyConfig](../service/#dependencyconfig)"
REF_DOCKER_CONFIG = "Refer to:<br>[Service - DockerConfig](../service/#dockerconfig)"
REF_ENVVAR = "Refer to:<br>[Service - Enviroment Variable](../service/#environmentvariable)"
REF_PV = "Refer to:<br>[Service - PeristentVolume](../service/#persistentvolume)"
REF_SERVICE = "Refer to:<br>[Service](../service/#service)"
REF_SUBMISSION_PARAMS = "Refer to:<br>[Service - SubmissionParams](../service/#submissionparams)"
REF_UPDATE_CONFIG = "Refer to:<br>[Service - UpdateConfig](../service/#updateconfig)"
REF_UPDATE_SOURCE = "Refer to:<br>[Service - UpdateSource](../service/#updatesource)"


@odm.model(index=False, store=False)
class EnvironmentVariable(odm.Model):
    name: str = odm.keyword(description=REF_ENVVAR)
    value: str = odm.keyword(description=REF_ENVVAR)


@odm.model(index=False, store=False, description="Docker Configuration Delta")
class DockerConfigDelta(odm.Model):
    allow_internet_access = odm.Optional(odm.Boolean(), description=REF_DOCKER_CONFIG)
    command = odm.Optional(odm.List(odm.Keyword()), description=REF_DOCKER_CONFIG)
    cpu_cores = odm.Optional(odm.Float(), description=REF_DOCKER_CONFIG)
    environment = odm.Optional(odm.List(odm.Compound(EnvironmentVariable)), description=REF_DOCKER_CONFIG)
    image = odm.Optional(odm.Keyword(), description=REF_DOCKER_CONFIG)
    registry_username = odm.Optional(odm.Keyword(default=""), description=REF_DOCKER_CONFIG)
    registry_password = odm.Optional(odm.Keyword(default=""), description=REF_DOCKER_CONFIG)
    registry_type = odm.Optional(odm.Enum(values=["docker", "harbor"]), description=REF_DOCKER_CONFIG)
    ports = odm.Optional(odm.List(odm.Keyword()), description=REF_DOCKER_CONFIG)
    ram_mb = odm.Optional(odm.Integer(), description=REF_DOCKER_CONFIG)
    ram_mb_min = odm.Optional(odm.Integer(), description=REF_DOCKER_CONFIG)
    service_account = odm.optional(odm.keyword(description=REF_DOCKER_CONFIG))
    labels = odm.Optional(odm.List(odm.Compound(EnvironmentVariable)), description=REF_DOCKER_CONFIG)


@odm.model(index=False, store=False)
class UpdateSourceDelta(odm.Model):
    name = odm.Optional(odm.Keyword(), description=REF_UPDATE_SOURCE)
    password = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    pattern = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    private_key = odm.Keyword(default="", description=REF_UPDATE_SOURCE)
    ca_cert = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    ssl_ignore_errors = odm.Boolean(default=False, description=REF_UPDATE_SOURCE)
    proxy = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    uri = odm.Optional(odm.Keyword(), description=REF_UPDATE_SOURCE)
    username = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    headers = odm.Optional(odm.List(odm.Compound(EnvironmentVariable)), description=REF_UPDATE_SOURCE)
    default_classification = odm.Optional(odm.Classification(), description=REF_UPDATE_SOURCE)
    git_branch = odm.Optional(odm.Keyword(default=""), description=REF_UPDATE_SOURCE)
    sync = odm.Optional(odm.Boolean(default=False), description=REF_UPDATE_SOURCE)
    fetch_method = odm.Optional(odm.Enum(values=FETCH_METHODS, default="GET"), description=REF_UPDATE_SOURCE)
    enabled = odm.Optional(odm.Boolean(default=True), description=REF_UPDATE_SOURCE)
    override_classification = odm.Optional(odm.Boolean(default=False), description=REF_UPDATE_SOURCE)
    configuration = odm.Optional(odm.Mapping(odm.Any(), default={}), description=REF_UPDATE_SOURCE)
    update_interval = odm.Optional(odm.Integer(min=0), description=REF_UPDATE_SOURCE)
    ignore_cache = odm.Optional(odm.Boolean(default=False), description=REF_UPDATE_SOURCE)
    data = odm.Optional(odm.Text(), description=REF_UPDATE_SOURCE)


@ odm.model(index=False, store=False)
class PersistentVolumeDelta(odm.Model):
    mount_path = odm.Optional(odm.Keyword(), description=REF_PV)
    capacity = odm.Optional(odm.Keyword(), description=REF_PV)
    storage_class = odm.Optional(odm.Keyword(), description=REF_PV)
    access_mode = odm.Optional(odm.Enum(values=['ReadWriteOnce', 'ReadWriteMany']), description=REF_PV)


@ odm.model(index=False, store=False)
class DependencyConfigDelta(odm.Model):
    container = odm.Optional(odm.Compound(DockerConfigDelta), description=REF_DEPENDENCY_CONFIG)
    volumes = odm.Mapping(odm.Compound(PersistentVolumeDelta), default={}, description=REF_DEPENDENCY_CONFIG)
    run_as_core: bool = odm.Optional(odm.Boolean(), description=REF_DEPENDENCY_CONFIG)


@ odm.model(index=False, store=False)
class UpdateConfigDelta(odm.Model):
    generates_signatures = odm.Optional(odm.Boolean(), index=True, description=REF_UPDATE_CONFIG)
    sources = odm.Optional(odm.List(odm.Compound(UpdateSourceDelta)), description=REF_UPDATE_CONFIG)
    update_interval_seconds = odm.Optional(odm.Integer(), description=REF_UPDATE_CONFIG)
    wait_for_update = odm.Optional(odm.Boolean(), description=REF_UPDATE_CONFIG)
    signature_delimiter = odm.Optional(odm.Enum(values=SIGNATURE_DELIMITERS.keys()), description=REF_UPDATE_CONFIG)
    custom_delimiter = odm.Optional(odm.Keyword(), description=REF_UPDATE_CONFIG)
    default_pattern = odm.Optional(odm.Text(), description=REF_UPDATE_CONFIG)


@ odm.model(index=False, store=False)
class SubmissionParamsDelta(odm.Model):
    default = odm.Optional(odm.Any(), description=REF_SUBMISSION_PARAMS)
    name = odm.Optional(odm.Keyword(), description=REF_SUBMISSION_PARAMS)
    type = odm.Optional(odm.Enum(values=['str', 'int', 'list', 'bool']), description=REF_SUBMISSION_PARAMS)
    value = odm.Optional(odm.Any(), description=REF_SUBMISSION_PARAMS)
    list = odm.Optional(odm.Any(), description=REF_SUBMISSION_PARAMS)
    hide = odm.Optional(odm.Boolean(), description=REF_SUBMISSION_PARAMS)


@ odm.model(index=True, store=False, description="Service Delta relative to Initial Service Configuration")
class ServiceDelta(odm.Model):
    accepts = odm.Optional(odm.Keyword(), store=True, description=REF_SERVICE)
    auto_update: bool | None = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    rejects = odm.Optional(odm.Keyword(), store=True, description=REF_SERVICE)
    category = odm.Optional(odm.Keyword(), store=True, copyto="__text__", description=REF_SERVICE)
    classification = odm.Optional(odm.ClassificationString(), description=REF_SERVICE)
    config = odm.Optional(odm.Mapping(odm.Any()), index=False, description=REF_SERVICE)
    description = odm.Optional(odm.Text(), store=True, copyto="__text__", description=REF_SERVICE)
    default_result_classification = odm.Optional(odm.ClassificationString(), description=REF_SERVICE)
    enabled = odm.Optional(odm.Boolean(), store=True, description=REF_SERVICE)
    is_external = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    licence_count = odm.Optional(odm.Integer(), description=REF_SERVICE)
    max_queue_length = odm.Optional(odm.Integer(), description=REF_SERVICE)
    min_instances = odm.Optional(odm.Integer(), description=REF_SERVICE)

    uses_tags: bool = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    uses_tag_scores: bool = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    uses_temp_submission_data: bool = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    uses_metadata: bool = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    monitored_keys = odm.optional(odm.sequence(odm.keyword()))

    name = odm.Optional(odm.Keyword(), store=True, copyto="__text__", description=REF_SERVICE)
    version = odm.Keyword(store=True, description=REF_SERVICE)

    privileged = odm.Optional(odm.Boolean(), description=REF_SERVICE)
    disable_cache = odm.Optional(odm.Boolean(), description=REF_SERVICE)

    stage = odm.Optional(odm.Keyword(), store=True, copyto="__text__", description=REF_SERVICE)
    submission_params = odm.Optional(odm.List(odm.Compound(SubmissionParamsDelta)), index=False,
                                     description=REF_SERVICE)
    timeout = odm.Optional(odm.Integer(), description=REF_SERVICE)

    docker_config: DockerConfigDelta = odm.Optional(odm.Compound(DockerConfigDelta), description=REF_SERVICE)
    dependencies: DependencyConfigDelta = odm.Mapping(odm.Compound(DependencyConfigDelta), default={},
                                                      description=REF_SERVICE)

    update_channel = odm.Optional(odm.Enum(values=["stable", "rc", "beta", "dev"]), description=REF_SERVICE)
    update_config: UpdateConfigDelta = odm.Optional(odm.Compound(UpdateConfigDelta), description=REF_SERVICE)

    recursion_prevention = odm.Optional(odm.sequence(odm.Keyword()), description=REF_SERVICE)



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


@odm.model(index=False, store=False, description="Submission Summary Model")
class SubmissionSummary(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED, description="Classification of the cache")
    filtered = odm.Boolean(default=False, description="Has this cache entry been filtered?")
    expiry_ts = odm.Date(index=True, description="Expiry timestamp")
    tags = odm.Text(description="Tags cache")
    attack_matrix = odm.Text(description="ATT&CK Matrix cache")
    heuristics = odm.Text(description="Heuristics cache")
    heuristic_sections = odm.Text(description="All sections mapping to the heuristics")
    heuristic_name_map = odm.Text(description="Map of heuristic names to IDs")



import os
import logging

from assemblyline.common import forge
from assemblyline.common.str_utils import safe_str
from assemblyline.common.uid import get_id_from_data
from assemblyline.odm.models.signature import Signature


class YaraImporter(object):
    def __init__(self, logger=None):
        if not logger:
            from assemblyline.common import log as al_log
            al_log.init_logging('yara_importer')
            logger = logging.getLogger('assemblyline.yara_importer')
            logger.setLevel(logging.INFO)

        self.ds = forge.get_datastore()
        self.classification = forge.get_classification()
        self.log = logger

    def get_signature_name(self, signature):
        name = None
        for line in signature.splitlines():
            line = line.strip()
            if line.startswith("rule ") or line.startswith("private rule ") \
                    or line.startswith("global rule ") or line.startswith("global private rule "):
                name = line.split(":")[0].split("{")[0]
                name = name.replace("global ", "").replace("private ", "").replace("rule ", "")
                break

        if name is None:
            return name
        return name.strip()

    def parse_meta(self, signature):
        meta = {}
        meta_started = False
        for line in signature.splitlines():
            line = line.strip()
            if not meta_started and line.startswith('meta') and line.endswith(':'):
                meta_started = True
                continue

            if meta_started:
                if line.startswith("//") or line == "":
                    continue

                if "=" not in line:
                    break

                key, val = line.split("=", 1)
                key = key.strip()
                val = val.strip().strip('"')
                meta[key] = safe_str(val)

        return meta

    def _save_signatures(self, signatures, source, default_status="TESTING"):
        saved_sigs = []
        order = 1
        for signature in signatures:
            signature_hash = get_id_from_data(signature, length=16)

            meta = self.parse_meta(signature)

            classification = meta.get('classification', self.classification.UNRESTRICTED)
            signature_id = meta.get('rule_id', meta.get('signature_id', meta.get('id' , signature_hash)))
            revision = meta.get('rule_version', meta.get('revision', 1))
            name = self.get_signature_name(signature)
            status = meta.get('al_status', default_status)

            key = f"yara_{source}_{signature_id}"

            sig = Signature({
                'classification': classification,
                "data": signature,
                "name": name,
                "order": order,
                "revision": int(revision),
                "signature_id": signature_id,
                "source": source,
                "status": status,
                "type": "yara"
            })
            self.ds.signature.save(key, sig)
            self.log.info("Added signature %s" % name)

            saved_sigs.append(sig)
            order += 1

        return saved_sigs

    def _split_signatures(self, data):
        current_signature = []
        signatures = []
        in_rule = False
        for line in data.splitlines():
            temp_line = line.strip()

            if in_rule:
                current_signature.append(line)

                if temp_line == "}":
                    signatures.append("\n".join(current_signature))
                    current_signature = []
                    in_rule = False

            if temp_line.startswith("rule ") or temp_line.startswith("private rule ") \
                    or temp_line.startswith("global rule ") or temp_line.startswith("global private rule "):
                in_rule = True
                current_signature.append(line)

        return signatures

    def import_data(self, yara_bin, source, default_status="TESTING"):
        return self._save_signatures(self._split_signatures(yara_bin), source, default_status=default_status)

    def import_file(self, cur_file, source=None, default_status="TESTING"):
        cur_file = os.path.expanduser(cur_file)
        if os.path.exists(cur_file):
            with open(cur_file, "r") as yara_file:
                yara_bin = yara_file.read()
                return self.import_data(yara_bin, source or os.path.basename(cur_file), default_status=default_status)
        else:
            raise Exception(f"File {cur_file} does not exists.")

    def import_files(self, files, default_status="TESTING"):
        output = {}
        for cur_file in files:
            output[cur_file] = self.import_file(cur_file, default_status=default_status)

        return output



import datetime
import rstr
import random
import time
from typing import Optional as _Optional, Dict, Any as _Any

from assemblyline.common.uid import get_random_id
from assemblyline.odm import Boolean, Enum, Keyword, Text, List, Model, Compound, Integer, Float, Date, Mapping, \
    Classification, ClassificationString, Optional, Any, forge, IP, Domain, MD5, SHA1, SHA256, PhoneNumber, MAC, \
    URIPath, URI, SSDeepHash, Email, Platform, Processor, UpperKeyword, Json, EmptyableKeyword, UNCPath, Long, \
    ValidatedKeyword
from assemblyline.odm.models.tagging import Tagging

config = forge.get_config()

ALPHA = "ABCDEFGHIJKLMNOPQRSTUPVXYZabcdefghijklmnopqrstuvwxyz"
HASH_ALPHA = "abcdef0123456789"
SSDEEP_ALPHA = f"{ALPHA}0123456789"
WORDS = """The Cyber Centre stays on the cutting edge of technology by working with commercial vendors of cyber security
technology to support their development of enhanced cyber defence tools To do this our experts survey the cyber
security market evaluate emerging technologies in order to determine their potential to improve cyber security
across the country The Cyber Centre supports innovation by collaborating with all levels of government private
industry academia to examine complex problems in cyber security We are constantly engaging partners to promote
an open innovative environment We invite partners to work with us but also promote other Government of Canada
innovation programs One of our key partnerships is with the Government of Canada Build in Canada Innovation Program
BCIP The BCIP helps Canadian companies of all sizes transition their state of the art goods services from the
laboratory to the marketplace For certain cyber security innovations the Cyber Centre performs the role of technical
authority We evaluate participating companies new technology provide feedback in order to assist them in bringing
their product to market To learn more about selling testing an innovation visit the BCIP website""".split()
WORDS = list(set(WORDS))
MAPPING_KEYS = ["key_a", "key_b", "key_c", "key_d", "key_e", "key_f"]
META_KEYS = ["ingest.name", "ingest.id", "date", "source", "file.original_ext", "file.original_name"]
EXT = [
    ".jpg",
    ".doc",
    ".exe",
    ".pdf",
    ".xls",
    ".lnk",
    ".gif",
    ".ppt"
]

DOM = [
    ".com",
    ".ca",
    ".biz",
    ".edu"
]

SERVICES = {
    "Beaver": ("External", "FILTER"),
    "NSRL": ("Filtering", "FILTER"),
    "PE": ("Static Analysis", "CORE"),
    "Extract": ("Extraction", "EXTRACT"),
    "TagCheck": ("Static Analysis", "POST"),
    "PDFId": ("Static Analysis", "CORE"),
    "PeePDF": ("Static Analysis", "CORE"),
    "Characterize": ("Static Analysis", "CORE"),
    "FrankenStrings": ("Static Analysis", "SECONDARY"),
    "APKaye": ("Static Analysis", "CORE"),
    "Cuckoo": ("Dynamic Analysis", "CORE"),
    "McAfee": ("Antivirus", "CORE"),
    "Metadefender": ("Antivirus", "CORE"),
    "Suricata": ("Networking", "SECONDARY")
}

LABELS = [
    "CAMPAIGN_1",
    "CAMPAIGN_2",
    "CAMPAIGN_3",
    "CAMPAIGN_4",
    "CAMPAIGN_5",
    "PHISHING",
    "SMTP",
    "HTTP",
    "COMPROMISED"
]

USERS = [
    "admin",
    "user"
]

GROUPS = ["USERS", "ADMINS", "ANALYSTS"]

F_TYPES = [
    "image/png",
    "executable/windows",
    "document/pdf",
    "document/office",
    "document/xml",
    "code/javascript",
    "code/vb"
]

RULES = [
    "BlackShades",
    "Punisher",
    "gh0st",
    "Xtreme",
    "Bozok",
    "CyberGate",
    "NanoCore",
    "xRAT",
    "VirusRat",
    "LuxNet",
    "njRat",
    "Pandora",
    "njrat",
    "darkcomet51",
    "PoisonIvy",
    "mraptor_oletools",
    "VBA_external_connections",
    "DarkComet",
    "darkcomet_rc4"
]


def get_random_rule() -> str:
    return f"YAR_SAMPLE.{random.choice(RULES)}"


def get_random_file_type() -> str:
    return random.choice(F_TYPES)


def get_random_word() -> str:
    return random.choice(WORDS)


def get_random_phrase(wmin: int = 2, wmax: int = 6) -> str:
    return " ".join([get_random_word() for _ in range(random.randint(wmin, wmax))])


def get_random_hash(hash_len: int) -> str:
    return "".join([random.choice(HASH_ALPHA) for _ in range(hash_len)])


def get_random_tlsh():
    return f"T{get_random_hash(71)}"


def get_random_heuristic_id() -> str:
    return f"{get_random_service_name().upper()}.{random.randint(1, 5)}"


def get_random_label() -> str:
    return random.choice(LABELS)


def get_random_user() -> str:
    return random.choice(USERS)


def get_random_groups() -> str:
    return random.choice(GROUPS)


def get_random_filename(smin: int = 1, smax: int = 3) -> str:
    return "_".join([get_random_word().lower() for _ in range(random.randint(smin, smax))]) + random.choice(EXT)


def get_random_directory(smin: int = 2, smax: int = 6) -> str:
    return "/".join([get_random_word().lower() for _ in range(random.randint(smin, smax))])


def get_random_string(smin: int = 4, smax: int = 24) -> str:
    return "".join([random.choice(ALPHA) for _ in range(random.randint(smin, smax))])


def get_random_email() -> str:
    return f"{get_random_word()}@{get_random_word()}{random.choice(DOM)}"


def get_random_host() -> str:
    return get_random_word().lower() + random.choice(DOM)


def get_random_service_name() -> str:
    return random.choice(list(SERVICES.keys()))


def get_random_service_version() -> str:
    return f"4.2.0.stable{random.randint(1, 4)}"


def get_random_ip() -> str:
    return ".".join([str(random.randint(1, 254)) for _ in range(4)])


def get_random_iso_date(epoch: _Optional[float] = None) -> str:
    if epoch is None:
        epoch = time.time() + random.randint(-3000000, 0)

    return datetime.datetime.fromtimestamp(epoch).isoformat() + "Z"


def get_random_mapping(field) -> Dict[str, _Any]:
    return {MAPPING_KEYS[i]: random_data_for_field(field, MAPPING_KEYS[i]) for i in range(random.randint(0, 5))}


def get_random_meta(field) -> Dict[str, _Any]:
    return {META_KEYS[i]: random_data_for_field(field, META_KEYS[i]) for i in range(random.randint(0, 5))}


def get_random_phone() -> str:
    return f'{random.choice(["", "+1 "])}{"-".join([str(random.randint(100, 999)) for _ in range(3)])}' \
        f'{str(random.randint(0, 9))}'


def get_random_mac() -> str:
    return ":".join([get_random_hash(2) for _ in range(6)])


def get_random_unc_path() -> str:
    return "\\\\" + '\\'.join([get_random_word() for _ in range(random.randint(2, 6))])


def get_random_uri_path() -> str:
    return f"/{'/'.join([get_random_word() for _ in range(random.randint(2, 6))])}"


def get_random_uri() -> str:
    return f"{random.choice(['http', 'https', 'ftp'])}://{get_random_host()}{get_random_uri_path()}"


def get_random_ssdeep() -> str:
    return f"{str(random.randint(30, 99999))}" \
        f":{''.join([random.choice(SSDEEP_ALPHA) for _ in range(random.randint(20, 64))])}" \
        f":{''.join([random.choice(SSDEEP_ALPHA) for _ in range(random.randint(20, 64))])}"


def get_random_platform() -> str:
    return f"{random.choice(['Windows', 'Linux', 'MacOS', 'Android', 'iOS'])}"


def get_random_processor() -> str:
    return f"{random.choice(['x86', 'x64'])}"


def get_random_tags() -> dict:
    desired_tag_types = [
        'attribution.actor',
        'network.static.ip',
        'network.dynamic.ip',
        'network.static.domain',
        'network.dynamic.domain',
        'network.static.uri',
        'network.dynamic.uri',
        'av.virus_name',
        'attribution.implant',
        'file.rule.yara',
        'file.behavior',
        'attribution.exploit',
        'technique.packer',
        'technique.obfuscation'
    ]
    out = {}
    flat_fields = Tagging.flat_fields()
    # noinspection PyUnresolvedReferences
    flat_fields.pop('file.rule')
    tag_list = random.choices(list(flat_fields.keys()), k=random.randint(0, 2))
    tag_list.extend(random.choices(desired_tag_types, k=random.randint(1, 2)))
    for key in tag_list:
        parts = key.split(".")
        d = out
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            d = d[part]

        if parts[-1] not in d:
            d[parts[-1]] = []

        for _ in range(random.randint(1, 2)):
            d[parts[-1]].append(random_data_for_field(flat_fields.get(key, Keyword()), key.split(".")[-1]))

    return out


# noinspection PyProtectedMember
def random_data_for_field(field, name: str, minimal: bool = False) -> _Any:
    if isinstance(field, Boolean):
        return random.choice([True, False])
    elif isinstance(field, Classification):
        if field.engine.enforce:
            possible_classifications = list(field.engine.list_all_classification_combinations(normalized=True))
            possible_classifications.extend([field.engine.UNRESTRICTED, field.engine.RESTRICTED])
        else:
            possible_classifications = [field.engine.UNRESTRICTED]
        return random.choice(possible_classifications)
    elif isinstance(field, ClassificationString):
        possible_classifications = [field.engine.UNRESTRICTED]
        return random.choice(possible_classifications)
    elif isinstance(field, Enum):
        return random.choice([x for x in field.values if x is not None])
    elif isinstance(field, List):
        return [random_data_for_field(field.child_type, name) if not isinstance(field.child_type, Model)
                else random_model_obj(field.child_type, as_json=True) for _ in range(random.randint(1, 4))]
    elif isinstance(field, Compound):
        if minimal:
            return random_minimal_obj(field.child_type, as_json=True)
        else:
            return random_model_obj(field.child_type, as_json=True)
    elif isinstance(field, Mapping):
        if name == 'metadata':
            return get_random_meta(field.child_type)
        else:
            return get_random_mapping(field.child_type)
    elif isinstance(field, Optional):
        if not minimal:
            return random_data_for_field(field.child_type, name)
        else:
            return field.child_type.default
    elif isinstance(field, Date):
        return get_random_iso_date()
    elif isinstance(field, (Integer, Long)):
        if name == 'depth':
            return random.randint(1, 3)
        return random.randint(field.min or 128, field.max or 4096)
    elif isinstance(field, Float):
        return random.randint(12800, 409600) / 100.0
    elif isinstance(field, MD5):
        return get_random_hash(32)
    elif isinstance(field, SHA1):
        return get_random_hash(40)
    elif isinstance(field, SHA256):
        return get_random_hash(64)
    elif isinstance(field, SSDeepHash):
        return get_random_ssdeep()
    elif isinstance(field, URI):
        return get_random_uri()
    elif isinstance(field, UNCPath):
        return get_random_unc_path()
    elif isinstance(field, URIPath):
        return get_random_uri_path()
    elif isinstance(field, MAC):
        return get_random_mac()
    elif isinstance(field, PhoneNumber):
        return get_random_phone()
    elif isinstance(field, IP):
        return get_random_ip()
    elif isinstance(field, Domain):
        return get_random_host()
    elif isinstance(field, Email):
        return get_random_email()
    elif isinstance(field, Platform):
        return get_random_platform()
    elif isinstance(field, Processor):
        return get_random_processor()
    elif isinstance(field, Json):
        return random.choice([
            get_random_word(),                                         # string
            random.choice([True, False]),                              # boolean
            [get_random_word() for _ in range(random.randint(2, 4))],  # list
            random.randint(0, 100)]                                    # number
        )
    elif isinstance(field, UpperKeyword):
        return get_random_word().upper()
    elif isinstance(field, ValidatedKeyword):
        # Generate value based on regex pattern
        return rstr.xeger(field.validation_regex)
    elif isinstance(field, Keyword) or isinstance(field, EmptyableKeyword):
        if name:
            if "sha256" in name:
                return get_random_hash(64)
            elif "tlsh" in name:
                return get_random_tlsh()
            elif "yara" in name:
                return get_random_rule()
            elif "filetype" in name:
                return get_random_file_type()
            elif "organisation" in name:
                return config.system.organisation
            elif "heur_id" in name:
                return get_random_heuristic_id()
            elif "label" in name:
                return get_random_label()
            elif "groups" in name:
                return get_random_groups()
            elif "owner" in name or "uname" in name or "submitter" in name or "created_by" in name:
                return get_random_user()
            elif "service_name" in name:
                return get_random_service_name()
            elif "service_version" in name:
                return get_random_service_version()
            elif "sid" in name or "ingest_id" in name or "workflow_id" in name:
                return get_random_id()
            elif "mac" in name:
                return get_random_hash(12).upper()
            elif "sha1" in name:
                return get_random_hash(40)
            elif "md5" in name or "scan_key" in name or "alert_id" in name:
                return get_random_hash(32)
            elif "host" in name or "node" in name or "domain" in name:
                return get_random_host()
            elif name.endswith("ip") or name.startswith("ip_"):
                return get_random_ip()
            elif "file" in name:
                return get_random_filename()
            elif "name" in name:
                return get_random_filename()
            elif "directory" in name:
                return get_random_directory()
            elif "version" in name:
                return get_random_word()

        return get_random_word()
    elif isinstance(field, Text):
        return get_random_phrase(wmin=4, wmax=40)
    elif isinstance(field, Any):
        return get_random_word()
    else:
        raise ValueError(f"Unknown field type {field.__class__}")


# noinspection PyProtectedMember
def random_model_obj(model, as_json: bool = False) -> _Any:
    if model == Tagging:
        data = get_random_tags()
        if as_json:
            return data
        else:
            return model(data)

    data = {}
    for f_name, f_value in model._odm_field_cache.items():
        data[f_name] = random_data_for_field(f_value, f_name)

    if as_json:
        return data
    else:
        return model(data)


# noinspection PyProtectedMember
def random_minimal_obj(model, as_json: bool = False) -> _Any:
    data = {}
    for f_name, f_value in model._odm_field_cache.items():
        if not f_value.default_set:
            data[f_name] = random_data_for_field(f_value, f_name, minimal=True)

    if as_json:
        return data
    else:
        return model(data)



from __future__ import annotations

from typing import Optional as Opt

from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.constants import (
    DEFAULT_SERVICE_ACCEPTS,
    DEFAULT_SERVICE_REJECTS,
)

Classification = forge.get_classification()

FETCH_METHODS = ["GET", "POST", "GIT"]

SIGNATURE_DELIMITERS = {
    'new_line': '\n',
    'double_new_line': '\n\n',
    'pipe': '|',
    'comma': ',',
    'space': ' ',
    'none': '',
    'file': '',
    'custom': ''
}


@odm.model(index=False, store=False, description="Environment Variable Model")
class EnvironmentVariable(odm.Model):
    name: str = odm.keyword(description="Name of Environment Variable")
    value: str = odm.keyword(description="Value of Environment Variable")


@odm.model(index=False, store=False, description="Docker Container Configuration")
class DockerConfig(odm.Model):
    allow_internet_access: bool = odm.Boolean(default=False, description="Does the container have internet-access?")
    command: Opt[list[str]] = odm.Optional(odm.List(odm.Keyword()),
                                           description="Command to run when container starts up.")
    cpu_cores: float = odm.Float(default=1.0, description="CPU allocation")
    environment: list[EnvironmentVariable] = odm.List(odm.Compound(EnvironmentVariable), default=[],
                                                      description="Additional environemnt variables for the container")
    image: str = odm.Keyword(description="Complete name of the Docker image with tag, may include registry")
    registry_username: Opt[str] = odm.Optional(odm.Keyword(default=""),
                                               description="The username to use when pulling the image")
    registry_password: Opt[str] = odm.Optional(odm.Keyword(default=""),
                                               description="The password or token to use when pulling the image")
    registry_type: str = odm.Enum(values=["docker", "harbor"], default='docker',
                                  description="The type of container registry")
    ports: list[str] = odm.List(odm.Keyword(), default=[], description="What ports of container to expose?")
    ram_mb: int = odm.Integer(default=512, description="Container RAM limit")
    ram_mb_min: int = odm.Integer(default=256, description="Container RAM request")
    service_account = odm.optional(odm.keyword(description="Service account to use for pods in kubernetes"))
    labels = odm.sequence(odm.compound(EnvironmentVariable), default=[], description="Additional container labels.")


@ odm.model(index=False, store=False, description="Container's Persistent Volume Configuration")
class PersistentVolume(odm.Model):
    mount_path = odm.Keyword(description="Path into the container to mount volume")
    capacity = odm.Keyword(description="The amount of storage allocated for volume")
    storage_class = odm.Keyword(description="Storage class used to create volume")
    access_mode = odm.Enum(default='ReadWriteOnce', values=['ReadWriteOnce', 'ReadWriteMany'],
                           description="Access mode for volume")


@ odm.model(index=False, store=False, description="Container's Dependency Configuration")
class DependencyConfig(odm.Model):
    container: DockerConfig = odm.Compound(DockerConfig, description="Docker container configuration for dependency")
    volumes = odm.Mapping(odm.Compound(PersistentVolume), default={}, description="Volume configuration for dependency")
    run_as_core: bool = odm.Boolean(default=False, description="Should this dependency run as other core components?")


@ odm.model(index=False, store=False, description="Update Source Configuration")
class UpdateSource(odm.Model):
    name: str = odm.Keyword(description="Name of source")
    password: Opt[str] = odm.Optional(odm.Keyword(default=""), description="Password used to authenticate with source")
    pattern: Opt[str] = odm.Optional(odm.Keyword(default=""),
                                     description="Pattern used to find files of interest from source")
    private_key: Opt[str] = odm.Optional(odm.Keyword(default=""),
                                         description="Private key used to authenticate with source")
    ca_cert: Opt[str] = odm.Optional(odm.Keyword(default=""), description="CA cert for source")
    ssl_ignore_errors: bool = odm.Boolean(default=False, description="Ignore SSL errors when reaching out to source?")
    proxy: Opt[str] = odm.Optional(odm.Keyword(default=""), description="Proxy server for source")
    uri: str = odm.Keyword(description="URI to source")
    username: Opt[str] = odm.Optional(odm.Keyword(default=""), description="Username used to authenticate with source")
    headers: list[EnvironmentVariable] = odm.List(odm.Compound(EnvironmentVariable), default=[], description="Headers")
    default_classification = odm.Classification(
        default=Classification.UNRESTRICTED,
        description="Default classification used in absence of one defined in files from source")
    git_branch: str = odm.Optional(odm.Keyword(default=""), description="Branch to checkout from Git repository.")
    sync: bool = odm.Boolean(default=False, description="Synchronize signatures with remote source. Allows system to auto-disable signatures no longer found in source.")
    fetch_method: str = odm.Enum(values=FETCH_METHODS, default="GET", description="Fetch method to be used with source")
    enabled: bool = odm.Boolean(default=True, description="Is this source active for periodic fetching?")
    override_classification: bool = odm.Boolean(default=False,
                                                description="Should the source's classfication override the signature's self-defined classification, if any?")
    configuration: dict = odm.Mapping(odm.Any(), default={}, index=False, store=False,
                                      description="Processing configuration for source")
    update_interval: Opt[int] = odm.Optional(odm.Integer(min=1),
                                             description="Update check interval, in seconds, for this source")
    ignore_cache: bool = odm.Boolean(default=False,
                                     description="Ignore source caching and forcefully fetch from source")
    data: str = odm.Optional(odm.Text(), index=False, store=False,
                             description="Data that's sent in a POST request (`fetch_method=\"POST\"`)")



@ odm.model(index=False, store=False, description="Update Configuration for Signatures")
class UpdateConfig(odm.Model):
    generates_signatures = odm.Boolean(index=True, default=False, description="Does the updater produce signatures?")
    sources = odm.List(odm.Compound(UpdateSource), default=[], description="List of external sources")
    update_interval_seconds = odm.Integer(description="Update check interval, in seconds")
    wait_for_update = odm.Boolean(default=False, description="Should the service wait for updates first?")
    signature_delimiter = odm.Enum(values=SIGNATURE_DELIMITERS.keys(),
                                   default="double_new_line",
                                   description="Delimiter used when given a list of signatures")
    custom_delimiter = odm.Optional(odm.Keyword(), description="Custom delimiter definition")
    default_pattern = odm.Text(default=".*", description="Default pattern used for matching files")


@ odm.model(index=False, store=False, description="Submission Parameters for Service")
class SubmissionParams(odm.Model):
    default = odm.Any(description="Default value (must match value in `value` field)")
    name = odm.Keyword(description="Name of parameter")
    type = odm.Enum(values=['str', 'int', 'list', 'bool'], description="Type of parameter")
    value = odm.Any(description="Default value (must match value in `default` field)")
    list = odm.Optional(odm.Any(), description="List of values if `type: list`")
    hide = odm.Boolean(default=False, description="Should this parameter be hidden?")


@ odm.model(index=True, store=False, description="Service Configuration")
class Service(odm.Model):
    # Regexes applied to assemblyline style file type string
    accepts = odm.Keyword(store=True, default=DEFAULT_SERVICE_ACCEPTS,
                          description="Regex to accept files as identified by Assemblyline")
    auto_update: bool | None = odm.Optional(odm.Boolean(), description="Should the service be auto-updated?")
    rejects = odm.Optional(odm.Keyword(store=True, default=DEFAULT_SERVICE_REJECTS),
                           description="Regex to reject files as identified by Assemblyline")

    category = odm.Keyword(store=True, default="Static Analysis", copyto="__text__",
                           description="Which category does this service belong to?")
    classification = odm.ClassificationString(
        default=Classification.UNRESTRICTED, description="Classification of the service"
    )
    config = odm.Mapping(odm.Any(), default={}, index=False, store=False, description="Service Configuration")
    description = odm.Text(store=True, default="NA", copyto="__text__", description="Description of service")
    default_result_classification = odm.ClassificationString(
        default=Classification.UNRESTRICTED, description="Default classification assigned to service results")
    enabled: bool = odm.Boolean(store=True, default=False, description="Is the service enabled (by default)?")
    is_external: bool = odm.Boolean(
        default=False, description="Does this service perform analysis outside of Assemblyline?")
    licence_count: int = odm.Integer(default=0, description="How many licences is the service allowed to use?")
    min_instances: int = odm.Optional(odm.Integer(), description="The minimum number of service instances. Overrides Scaler's min_instances configuration.")
    max_queue_length: int = odm.Integer(
        default=0,
        description="If more than this many jobs are queued for this service drop those over this limit. 0 is unlimited.")

    uses_tags: bool = odm.Boolean(
        default=False, description="Does this service use tags from other services for analysis?")
    uses_tag_scores: bool = odm.Boolean(
        default=False, description="Does this service use scores of tags from other services for analysis?")
    uses_temp_submission_data: bool = odm.Boolean(
        default=False, description="Does this service use temp data from other services for analysis?")
    uses_metadata: bool = odm.Boolean(
        default=False, description="Does this service use submission metadata for analysis?")
    monitored_keys: list[str] = odm.sequence(
        odm.keyword(), default=[],
        description="This service watches these temporary keys for changes when partial results are produced.")

    name: str = odm.Keyword(store=True, copyto="__text__", description="Name of service")
    version = odm.Keyword(store=True, description="Version of service")

    privileged = odm.Boolean(
        default=False,
        description="Should the service be able to talk to core infrastructure or just service-server for tasking?")
    disable_cache = odm.Boolean(default=False, description="Should the result cache be disabled for this service?")

    stage = odm.Keyword(store=True, default="CORE", copyto="__text__",
                        description="Which execution stage does this service run in?")
    submission_params: list[SubmissionParams] = odm.sequence(
        odm.compound(SubmissionParams), index=False, default=[],
        description="Submission parameters of service")
    timeout: int = odm.Integer(default=60, description="Service task timeout, in seconds")

    docker_config: DockerConfig = odm.Compound(DockerConfig, description="Docker configuration for service")
    dependencies: dict[str, DependencyConfig] = odm.Mapping(odm.Compound(
        DependencyConfig), default={}, description="Dependency configuration for service")

    update_channel: str = odm.Enum(
        values=["stable", "rc", "beta", "dev"],
        default='stable', description="What channel to watch for service updates?")
    update_config: UpdateConfig = odm.Optional(
        odm.Compound(UpdateConfig),
        description="Update configuration for fetching external resources")

    recursion_prevention: list[str] = odm.sequence(odm.Keyword(), default=[], description="List of service names/categories where recursion is prevented.")



from assemblyline import odm
from assemblyline.common import forge, constants
from assemblyline.odm.models.submission import ServiceSelection

Classification = forge.get_classification()

ENCODINGS = {"cart", "raw", "zip"}
VIEWS = {"report", "details"}


@odm.model(index=False, store=False, description="Model of User Settings")
class UserSettings(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED,
                                        description="Default submission classification")
    deep_scan = odm.Boolean(default=False, description="Should a deep scan be performed?")
    description = odm.Keyword(default="", description="Default description")
    download_encoding = odm.Enum(values=ENCODINGS, default="cart",
                                 description="Default download encoding when downloading files")
    default_external_sources = odm.List(odm.Keyword(), default=[],
                                        description="List of sha256 sources to check by default")
    default_zip_password = odm.Text(
        default="infected",
        description="Default user-defined password for creating password protected ZIPs when downloading files"
    )
    executive_summary = odm.Boolean(default=True, description="Should executive summary sections be shown?")
    expand_min_score = odm.Integer(default=500, description="Auto-expand section when score bigger then this")
    generate_alert = odm.Boolean(default=False, description="Generate an alert?")
    ignore_cache = odm.Boolean(default=False, description="Ignore service caching?")

    #the following 1 line can be removed after assemblyline 4.6+
    ignore_dynamic_recursion_prevention = odm.Boolean(default=False, description="Ignore dynamic recursion prevention?")
    ignore_recursion_prevention = odm.Boolean(default=False, description="Ignore all service recursion prevention?")
    ignore_filtering = odm.Boolean(default=False, description="Ignore filtering services?")
    malicious = odm.Boolean(default=False, description="Is the file submitted already known to be malicious?")
    priority = odm.Integer(default=1000, description="Default priority for the submissions",
                           min=1, max=constants.MAX_PRIORITY)
    profile = odm.Boolean(default=False, description="Should the submission do extra profiling?")
    service_spec = odm.Mapping(odm.Mapping(odm.Any()), default={}, description="Default service specific settings")
    services = odm.Compound(ServiceSelection, default={}, description="Default service selection")
    submission_view = odm.Enum(values=VIEWS, default="report", description="Default view for completed submissions")
    ttl = odm.Integer(default=30, description="Default submission TTL, in days")



from assemblyline import odm
from assemblyline.odm.messages import PerformanceTimer

MSG_TYPES = {"ServiceTimingHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.service_heartbeat.ServiceTimingMessage"


@odm.model(description="Timing Metrics")
class Metrics(odm.Model):
    execution = PerformanceTimer(description="Excution time")
    execution_count = odm.Integer(description="Number of executes")
    idle = PerformanceTimer(description="Idle time")
    idle_count = odm.Integer(description="Number of idles")


@odm.model(description="Hearbeat Model")
class Heartbeat(odm.Model):
    instances = odm.Integer(description="Number of instances")
    metrics = odm.Compound(Metrics, description="Metrics")
    queue = odm.Integer(description="Queue size")
    service_name = odm.Keyword(description="Name of service")


@odm.model(description="Model of Service Timing Heartbeat Message")
class ServiceTimingMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="ServiceTimingHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from collections import defaultdict
from typing import Any, Dict

from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.caching import generate_conf_key
from assemblyline.common.dict_utils import flatten
from assemblyline.common.tagging import tag_dict_to_list
from assemblyline.odm.models.tagging import Tagging

# This is a "keys-only" representation of the BODY_FORMAT StringTable in
# assemblyline-v4-service/assemblyline_v4_service/common/result.py.
# Any updates here need to go in that StringTable also.
BODY_FORMAT = {
    "TEXT",
    "MEMORY_DUMP",
    "GRAPH_DATA",
    "URL",
    "JSON",
    "KEY_VALUE",
    "PROCESS_TREE",
    "TABLE",
    "IMAGE",
    "MULTI",
    "ORDERED_KEY_VALUE",
    "TIMELINE"
}
# This is a "keys-only" representation of the PROMOTE_TO StringTable in
# assemblyline-v4-service/assemblyline_v4_service/common/result.py.
# Any updates here need to go in that StringTable also.
PROMOTE_TO = {"SCREENSHOT", "ENTROPY", "URI_PARAMS"}
constants = forge.get_constants()


@odm.model(index=True, store=False)
class Attack(odm.Model):
    attack_id = odm.Keyword(copyto="__text__", description="ID", ai=False)
    pattern = odm.Keyword(copyto="__text__", description="Pattern Name")
    categories = odm.List(odm.Keyword(), description="Categories")


@odm.model(index=True, store=False, description="Heuristic Signatures")
class Signature(odm.Model):
    name = odm.Keyword(copyto="__text__", description="Name of the signature that triggered the heuristic")
    frequency = odm.Integer(default=1, description="Number of times this signature triggered the heuristic")
    safe = odm.Boolean(default=False, description="Is the signature safelisted or not")


@odm.model(index=True, store=False, description="Heuristic associated to the Section")
class Heuristic(odm.Model):
    heur_id = odm.Keyword(copyto="__text__", description="ID of the heuristic triggered", ai=False)
    name = odm.Keyword(copyto="__text__", description="Name of the heuristic")
    attack = odm.List(odm.Compound(Attack), default=[], description="List of Att&ck IDs related to this heuristic")
    signature = odm.List(odm.Compound(Signature), default=[],
                         description="List of signatures that triggered the heuristic", ai=False)
    score = odm.Integer(description="Calculated Heuristic score")


@odm.model(index=True, store=False, description="Result Section")
class Section(odm.Model):
    auto_collapse = odm.Boolean(default=False, description="Should the section be collapsed when displayed?", ai=False)
    body = odm.Optional(odm.Text(copyto="__text__"), description="Text body of the result section")
    classification = odm.Classification(description="Classification of the section", ai=False)
    body_format = odm.Enum(values=BODY_FORMAT, index=False, description="Type of body in this section")
    body_config = odm.Optional(odm.Mapping(odm.Any(), index=False,
                               description="Configurations for the body of this section"), ai=False)
    depth = odm.Integer(index=False, description="Depth of the section", ai=False)
    heuristic = odm.Optional(odm.Compound(Heuristic), description="Heuristic used to score result section")
    tags = odm.Compound(Tagging, default={}, description="List of tags associated to this section")
    safelisted_tags = odm.FlattenedListObject(store=False, default={}, description="List of safelisted tags", ai=False)
    title_text = odm.Text(copyto="__text__", description="Title of the section")
    promote_to = odm.Optional(odm.Enum(
        values=PROMOTE_TO,
        description="This is the type of data that the current section should be promoted to.", ai=False))


@odm.model(index=True, store=True, description="Result Body")
class ResultBody(odm.Model):
    score = odm.Integer(default=0, description="Aggregate of the score for all heuristics")
    sections = odm.List(odm.Compound(Section), default=[], description="List of sections")


@odm.model(index=False, store=False, description="Service Milestones")
class Milestone(odm.Model):
    service_started = odm.Date(default="NOW", description="Date the service started scanning")
    service_completed = odm.Date(default="NOW", description="Date the service finished scanning")


@odm.model(index=True, store=False, description="File related to the Response")
class File(odm.Model):
    name = odm.Keyword(copyto="__text__", description="Name of the file")
    sha256 = odm.SHA256(copyto="__text__", description="SHA256 of the file")
    description = odm.Text(copyto="__text__", description="Description of the file")
    classification = odm.Classification(description="Classification of the file", ai=False)
    is_section_image = odm.Boolean(default=False,
                                   description="Is this an image used in an Image Result Section?", ai=False)
    # Possible values for PARENT_RELATION can be found in
    # assemblyline-v4-service/assemblyline_v4_service/common/task.py.
    parent_relation = odm.Text(
        default="EXTRACTED",
        description="File relation to parent, if any.\
            <br>Values: `\"ROOT\", \"EXTRACTED\", \"INFORMATION\", \"DYNAMIC\", \"MEMDUMP\", \"DOWNLOADED\"`", ai=False
    )
    allow_dynamic_recursion = odm.Boolean(
        default=False,
        description="Allow file to be analysed during Dynamic Analysis"
                    "even if Dynamic Recursion Prevention is enabled.", ai=False)


@odm.model(index=True, store=True, description="Response Body of Result")
class ResponseBody(odm.Model):
    milestones = odm.Compound(Milestone, default={}, description="Milestone block", ai=False)
    service_version = odm.Keyword(store=False, description="Version of the service", ai=False)
    service_name = odm.Keyword(copyto="__text__", description="Name of the service that scanned the file")
    service_tool_version = odm.Optional(
        odm.Keyword(copyto="__text__"),
        description="Tool version of the service", ai=False)
    supplementary = odm.List(odm.Compound(File), default=[], description="List of supplementary files", ai=False)
    extracted = odm.List(odm.Compound(File), default=[], description="List of extracted files")
    service_context = odm.Optional(
        odm.Keyword(index=False, store=False),
        description="Context about the service", ai=False)
    service_debug_info = odm.Optional(
        odm.Keyword(index=False, store=False),
        description="Debug info about the service", ai=False)


@odm.model(index=True, store=True, description="Result Model")
class Result(odm.Model):
    archive_ts = odm.Optional(odm.Date(description="Time at which the result was archived", ai=False))
    classification = odm.Classification(description="Aggregate classification for the result", ai=False)
    created = odm.Date(default="NOW", description="Date at which the result object got created", ai=False)
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry timestamp", ai=False)
    response: ResponseBody = odm.compound(ResponseBody, description="The body of the response from the service")
    result: ResultBody = odm.compound(ResultBody, default={}, description="The result body")
    sha256 = odm.SHA256(store=False, description="SHA256 of the file the result object relates to")
    type = odm.Optional(odm.Keyword())
    size = odm.Optional(odm.Integer())
    drop_file = odm.Boolean(default=False, description="Use to not pass to other stages after this run", ai=False)
    partial = odm.Boolean(default=False, description="Invalidate the current result cache creation", ai=False)
    from_archive = odm.Boolean(index=False, default=False, description="Was loaded from the archive", ai=False)

    def build_key(self, service_tool_version=None, task=None):
        return self.help_build_key(
            self.sha256,
            self.response.service_name,
            self.response.service_version,
            self.is_empty(),
            service_tool_version=service_tool_version,
            task=task,
            partial=self.partial
        )

    @staticmethod
    def help_build_key(sha256, service_name, service_version, is_empty, service_tool_version=None, task=None, partial=False):
        key_list = [
            sha256,
            service_name.replace('.', '_'),
            f"v{service_version.replace('.', '_')}",
            f"c{generate_conf_key(service_tool_version=service_tool_version, task=task, partial_result=partial)}",
        ]

        if is_empty:
            key_list.append("e")

        return '.'.join(key_list)

    def scored_tag_dict(self) -> Dict[str, Dict[str, Any]]:
        tags: Dict[str, Dict[str, Any]] = defaultdict(lambda: {'score': 0})
        # Save the tags and their score
        for section in self.result.sections:
            tag_list = tag_dict_to_list(flatten(section.tags.as_primitives()))
            for tag in tag_list:
                key = f"{tag['type']}:{tag['value']}"
                tags[key].update(tag)
                tags[key]['score'] += section.heuristic.score if section.heuristic else 0

        return tags

    def is_empty(self) -> bool:
        if len(self.response.extracted) == 0 and \
                len(self.response.supplementary) == 0 and \
                len(self.result.sections) == 0 and \
                self.result.score == 0 and not self.partial:
            return True
        return False



import logging
import logging.config
import logging.handlers

from traceback import format_exception
from typing import Optional

import json
import os

from assemblyline.common import forge
from assemblyline.common.logformat import AL_LOG_FORMAT, AL_SYSLOG_FORMAT, AL_JSON_FORMAT
from assemblyline.odm.models.config import Config

# Check to see if a log level override is set in the environment.
# This is useful for debugging purposes.
LOG_LEVEL_OVERRIDE = os.environ.get("LOG_LEVEL")

log_level_map = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "DISABLED": 60
}


class PrintLogger(object):
    @staticmethod
    def info(msg, end=None):
        print(msg, end=end)

    @staticmethod
    def warning(msg, end=None):
        print(f"[W] {msg}", end=end)

    @staticmethod
    def warn(msg, end=None):
        print(f"[W] {msg}", end=end)

    @staticmethod
    def error(msg, end=None):
        print(f"[E] {msg}", end=end)

    @staticmethod
    def exception(msg, end=None):
        print(f"[EX] {msg}", end=end)


class JsonFormatter(logging.Formatter):
    def formatMessage(self, record):
        if record.exc_info:
            record.exc_text = self.formatException(record.exc_info)
            record.exc_info = None

        if record.exc_text:
            record.message += '\n' + record.exc_text
            record.exc_text = None

        record.message = json.dumps(record.message)
        return self._style.format(record)

    def formatException(self, exc_info):
        return ''.join(format_exception(*exc_info))


def init_logging(name: str, config: Config = None, log_level: Optional[str] = None):
    logger = logging.getLogger('assemblyline')

    # If the environment has a log level override, use it.
    if not log_level and LOG_LEVEL_OVERRIDE:
        log_level = LOG_LEVEL_OVERRIDE

    # Test if we've initialized the log handler already.
    if len(logger.handlers) != 0:
        return

    if name.startswith("assemblyline."):
        name = name[13:]

    if config is None:
        config = forge.get_config()

    if log_level is None:
        log_level = log_level_map[config.logging.log_level]

    logging.root.setLevel(logging.CRITICAL)
    logger.setLevel(log_level)

    if config.logging.log_level == "DISABLED":
        # While log_level is set to disable, we will not create any handlers
        return

    if config.logging.log_to_file:
        if not os.path.isdir(config.logging.log_directory):
            print('Warning: log directory does not exist. Will try to create %s' % config.logging.log_directory)
            os.makedirs(config.logging.log_directory)

        if log_level <= logging.DEBUG:
            dbg_file_handler = logging.handlers.RotatingFileHandler(
                os.path.join(config.logging.log_directory, f'{name}.dbg'), maxBytes=10485760, backupCount=5)
            dbg_file_handler.setLevel(logging.DEBUG)
            if config.logging.log_as_json:
                dbg_file_handler.setFormatter(JsonFormatter(AL_JSON_FORMAT))
            else:
                dbg_file_handler.setFormatter(logging.Formatter(AL_LOG_FORMAT))
            logger.addHandler(dbg_file_handler)

        if log_level <= logging.INFO:
            op_file_handler = logging.handlers.RotatingFileHandler(
                os.path.join(config.logging.log_directory, f'{name}.log'), maxBytes=10485760, backupCount=5)
            op_file_handler.setLevel(logging.INFO)
            if config.logging.log_as_json:
                op_file_handler.setFormatter(JsonFormatter(AL_JSON_FORMAT))
            else:
                op_file_handler.setFormatter(logging.Formatter(AL_LOG_FORMAT))
            logger.addHandler(op_file_handler)

        if log_level <= logging.ERROR:
            err_file_handler = logging.handlers.RotatingFileHandler(
                os.path.join(config.logging.log_directory, f'{name}.err'), maxBytes=10485760, backupCount=5)
            err_file_handler.setLevel(logging.ERROR)
            if config.logging.log_as_json:
                err_file_handler.setFormatter(JsonFormatter(AL_JSON_FORMAT))
            else:
                err_file_handler.setFormatter(logging.Formatter(AL_LOG_FORMAT))
            err_file_handler.setFormatter(logging.Formatter(AL_LOG_FORMAT))
            logger.addHandler(err_file_handler)

    if config.logging.log_to_console:
        console = logging.StreamHandler()
        if config.logging.log_as_json:
            console.setFormatter(JsonFormatter(AL_JSON_FORMAT))
        else:
            console.setFormatter(logging.Formatter(AL_LOG_FORMAT))
        logger.addHandler(console)

    if config.logging.log_to_syslog and config.logging.syslog_host and config.logging.syslog_port:
        syslog_handler = logging.handlers.SysLogHandler(address=(config.logging.syslog_host,
                                                                 config.logging.syslog_port))
        syslog_handler.formatter = logging.Formatter(AL_SYSLOG_FORMAT)
        logger.addHandler(syslog_handler)
